@@ -1,6 +1,6 @@
 /* linux.c -- Linux specific code for ifconfig
   Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
-  2010, 2011, 2012 Free Software Foundation, Inc.
+  2010, 2011, 2012, 2013, 2014, 2015 Free Software Foundation, Inc.
 
   This file is part of GNU Inetutils.
 
@@ -45,6 +45,7 @@
 #include <linux/if_ether.h>
 
 #include <read-file.h>
+#include <unused-parameter.h>
 
 #include "../ifconfig.h"
 
@@ -52,22 +53,28 @@
 /* ARPHRD stuff.  */
 
 static void
-print_hwaddr_ether (format_data_t form, unsigned char *data)
+print_hwaddr_ether (format_data_t form _GL_UNUSED_PARAMETER,
+		    unsigned char *data)
 {
   *column += printf ("%02X:%02X:%02X:%02X:%02X:%02X",
 		     data[0], data[1], data[2], data[3], data[4], data[5]);
+  had_output = 1;
 }
 
 static void
-print_hwaddr_arcnet (format_data_t form, unsigned char *data)
+print_hwaddr_arcnet (format_data_t form _GL_UNUSED_PARAMETER,
+		     unsigned char *data)
 {
   *column += printf ("%02X", data[0]);
+  had_output = 1;
 }
 
 static void
-print_hwaddr_dlci (format_data_t form, unsigned char *data)
+print_hwaddr_dlci (format_data_t form _GL_UNUSED_PARAMETER,
+		   unsigned char *data)
 {
   *column += printf ("%i", *((short *) data));
+  had_output = 1;
 }
 
 static void
@@ -81,22 +88,29 @@ print_hwaddr_ax25 (format_data_t form, unsigned char *data)
     }
   i = (data[6] & 0x1E) >> 1;
   if (i)
-    *column += printf ("-%i", i);
+    {
+      *column += printf ("-%i", i);
+      had_output = 1;
+    }
 #undef mangle
 }
 
 static void
-print_hwaddr_irda (format_data_t form, unsigned char *data)
+print_hwaddr_irda (format_data_t form _GL_UNUSED_PARAMETER,
+		   unsigned char *data)
 {
   *column += printf ("%02X:%02X:%02X:%02X",
 		     data[3], data[2], data[1], data[0]);
+  had_output = 1;
 }
 
 static void
-print_hwaddr_rose (format_data_t form, unsigned char *data)
+print_hwaddr_rose (format_data_t form _GL_UNUSED_PARAMETER,
+		   unsigned char *data)
 {
   *column += printf ("%02X%02X%02X%02X%02X",
 		     data[0], data[1], data[2], data[3], data[4]);
+  had_output = 1;
 }
 
 struct arphrd_symbol
@@ -419,7 +433,7 @@ pnd_read ()
 
   while (getline (&buf, &bufsize, fp) > 0)
     {
-      struct pnd_stats *stats = xzalloc (sizeof (*stats));;
+      struct pnd_stats *stats = xzalloc (sizeof (*stats));
       char *p = buf, *q;
       size_t len;
 
@@ -535,7 +549,8 @@ system_fh_hwaddr_query (format_data_t form, int argc, char *argv[])
 }
 
 void
-system_fh_hwaddr (format_data_t form, int argc, char *argv[])
+system_fh_hwaddr (format_data_t form, int argc _GL_UNUSED_PARAMETER,
+		  char *argv[] _GL_UNUSED_PARAMETER)
 {
 #ifdef SIOCGIFHWADDR
   if (ioctl (form->sfd, SIOCGIFHWADDR, form->ifr) < 0)
@@ -571,7 +586,8 @@ system_fh_hwtype_query (format_data_t form, int argc, char *argv[])
 }
 
 void
-system_fh_hwtype (format_data_t form, int argc, char *argv[])
+system_fh_hwtype (format_data_t form, int argc _GL_UNUSED_PARAMETER,
+		  char *argv[] _GL_UNUSED_PARAMETER)
 {
 #ifdef SIOCGIFHWADDR
   if (ioctl (form->sfd, SIOCGIFHWADDR, form->ifr) < 0)
@@ -594,6 +610,38 @@ system_fh_hwtype (format_data_t form, int argc, char *argv[])
 #endif
 }
 
+/* Accept every value of metric as printable.  The net-tools'
+ * implementation of ifconfig displays metric 0 as `1', so we
+ * aim at the same thing, even though all other unices disagree.
+ */
+void
+system_fh_metric_query (format_data_t form, int argc, char *argv[])
+{
+#ifdef SIOCGIFMETRIC
+  if (ioctl (form->sfd, SIOCGIFMETRIC, form->ifr) >= 0)
+    select_arg (form, argc, argv, 0);
+  else
+#endif
+    select_arg (form, argc, argv, 1);
+}
+
+void
+system_fh_metric (format_data_t form, int argc, char *argv[])
+{
+#ifdef SIOCGIFMETRIC
+  if (ioctl (form->sfd, SIOCGIFMETRIC, form->ifr) < 0)
+    error (EXIT_FAILURE, errno,
+	   "SIOCGIFMETRIC failed for interface `%s'",
+	   form->ifr->ifr_name);
+  else
+    put_int (form, argc, argv,
+	     form->ifr->ifr_metric ? form->ifr->ifr_metric : 1);
+#else
+  *column += printf ("(not available)");
+  had_output = 1;
+#endif
+}
+
 void
 system_fh_txqlen_query (format_data_t form, int argc, char *argv[])
 {
@@ -661,8 +709,8 @@ system_parse_opt_set_txqlen (struct ifconfig *ifp, char *arg)
 
 static struct argp_option linux_argp_options[] = {
   { "txqlen", 'T', "N", 0,
-    "set transmit queue length to N" },
-  { NULL }
+    "set transmit queue length to N", 0 },
+  { NULL, 0, NULL, 0, NULL, 0 }
 };
 
 static error_t
@@ -681,7 +729,8 @@ linux_argp_parser (int key, char *arg, struct argp_state *state)
   return 0;
 }
 
-static struct argp linux_argp = { linux_argp_options, linux_argp_parser };
+static struct argp linux_argp =
+  { linux_argp_options, linux_argp_parser, NULL, NULL, NULL, NULL, NULL };
 
 struct argp_child system_argp_child = {
   &linux_argp,
@@ -714,7 +763,7 @@ system_parse_opt_rest (struct ifconfig **ifp, int argc, char *argv[])
   int i = 0;
   enum
   {
-    EXPECT_INET,
+    EXPECT_AF,
     EXPECT_NOTHING,
     EXPECT_BROADCAST,
     EXPECT_DSTADDR,
@@ -722,7 +771,7 @@ system_parse_opt_rest (struct ifconfig **ifp, int argc, char *argv[])
     EXPECT_MTU,
     EXPECT_METRIC,
     EXPECT_TXQLEN,
-  } expect = EXPECT_INET;
+  } expect = EXPECT_AF;
   int mask, rev;
 
   *ifp = parse_opt_new_ifs (argv[0]);
@@ -755,10 +804,15 @@ system_parse_opt_rest (struct ifconfig **ifp, int argc, char *argv[])
 	  system_parse_opt_set_txqlen (*ifp, argv[i]);
 	  break;
 
-	case EXPECT_INET:
+	case EXPECT_AF:
 	  expect = EXPECT_NOTHING;
 	  if (!strcmp (argv[i], "inet"))
 	    continue;
+	  else if (!strcmp (argv[i], "inet6"))
+	    {
+	      error (0, 0, "%s is not a supported address family", argv[i]);
+	      return 0;
+	    }
 	  break;
 
 	case EXPECT_NOTHING:
@@ -784,7 +838,8 @@ system_parse_opt_rest (struct ifconfig **ifp, int argc, char *argv[])
 	parse_opt_set_flag (*ifp, IFF_UP | IFF_RUNNING, 0);
       else if (!strcmp (argv[i], "down"))
 	parse_opt_set_flag (*ifp, IFF_UP, 1);
-      else if ((mask = if_nameztoflag (argv[i], &rev)) != 0)
+      else if (((mask = if_nameztoflag (argv[i], &rev))
+		& ~IU_IFF_CANTCHANGE) != 0)
 	parse_opt_set_flag (*ifp, mask, rev);
       else
 	parse_opt_set_address (*ifp, argv[i]);
@@ -816,7 +871,7 @@ system_parse_opt_rest (struct ifconfig **ifp, int argc, char *argv[])
       error (0, 0, "option `txqueuelen' requires an argument");
       break;
 
-    case EXPECT_INET:
+    case EXPECT_AF:
     case EXPECT_NOTHING:
       return 1;
     }
