@@ -1,6 +1,6 @@
 /*
   Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008,
-  2009, 2010, 2011, 2012 Free Software Foundation, Inc.
+  2009, 2010, 2011, 2012, 2013, 2014, 2015 Free Software Foundation, Inc.
 
   This file is part of GNU Inetutils.
 
@@ -55,9 +55,23 @@
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
+#include <unused-parameter.h>
 
 #include "fts.h"
 
+/* Largest alignment size needed, minus one.
+ * It is used as value and as bit mask.
+ * Usually long double is the worst case.  */
+#ifndef ALIGNBYTES
+# define ALIGNBYTES	(__alignof__ (long double) - 1)
+#endif
+
+/* Align P to size ALIGNBYTES at adapted boundary.  */
+#ifndef ALIGN
+# define ALIGN(p)	(((unsigned long int) (p) + ALIGNBYTES) & ~ALIGNBYTES)
+#endif
+
+
 static FTSENT *fts_alloc (FTS *, const char *, int);
 static FTSENT *fts_build (FTS *, int);
 static void fts_lfree (FTSENT *);
@@ -512,7 +526,7 @@ next:tmp = p;
  * reasons.
  */
 int
-fts_set (FTS *sp, FTSENT *p, int instr)
+fts_set (FTS *sp _GL_UNUSED_PARAMETER, FTSENT *p, int instr)
 {
   if (instr && instr != FTS_AGAIN && instr != FTS_FOLLOW &&
       instr != FTS_NOINSTR && instr != FTS_SKIP)
@@ -618,7 +632,7 @@ fts_build (register FTS *sp, int type)
   void *adjaddr;
   int cderrno, descend, len, level, maxlen, nlinks, saved_errno;
   char *cp = NULL;
-#ifdef DTF_HIDEW
+#ifdef HAVE___OPENDIR2
   int oflag;
 #endif
 
@@ -629,15 +643,24 @@ fts_build (register FTS *sp, int type)
    * Open the directory for reading.  If this fails, we're done.
    * If being called from fts_read, set the fts_info field.
    */
-#if defined HAVE_OPENDIR2 && defined DTF_HIDEW
-  if (ISSET (FTS_WHITEOUT))
-    oflag = DTF_NODUP | DTF_REWIND;
-  else
-    oflag = DTF_HIDEW | DTF_NODUP | DTF_REWIND;
-#else
-# define opendir2(path, flag) opendir(path)
-#endif
-  if ((dirp = opendir2 (cur->fts_accpath, oflag)) == NULL)
+#if defined HAVE___OPENDIR2
+  oflag = DTF_NODUP;
+
+# ifdef DTF_REWIND
+  oflag |= DTF_REWIND;
+# endif
+
+# ifdef DTF_HIDEW
+  if (!ISSET (FTS_WHITEOUT))
+    oflag |= DTF_HIDEW;
+# endif /* DTF_HIDEW */
+
+  dirp = __opendir2 (cur->fts_accpath, oflag);
+#else /* !HAVE___OPENDIR2 */
+  dirp = opendir (cur->fts_accpath);
+#endif /* !HAVE___OPENDIR2 */
+
+  if (dirp == NULL)
     {
       if (type == BREAD)
 	{
@@ -990,7 +1013,7 @@ fts_alloc (FTS *sp, const char *name, register int namelen)
    */
   len = sizeof (FTSENT) + namelen;
   if (!ISSET (FTS_NOSTAT))
-    len += sizeof (struct stat);
+    len += sizeof (struct stat) + ALIGNBYTES;
   if ((p = malloc (len)) == NULL)
     return (NULL);
 
@@ -998,7 +1021,7 @@ fts_alloc (FTS *sp, const char *name, register int namelen)
   memmove (p->fts_name, name, namelen + 1);
 
   if (!ISSET (FTS_NOSTAT))
-    p->fts_statp = (struct stat *) (p->fts_name + namelen + 2);
+    p->fts_statp = (struct stat *) ALIGN (p->fts_name + namelen + 2);
   p->fts_namelen = namelen;
   p->fts_path = sp->fts_path;
   p->fts_errno = 0;
