@@ -1,7 +1,7 @@
 /*
   Copyright (C) 1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003,
-  2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013 Free
-  Software Foundation, Inc.
+  2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014,
+  2015 Free Software Foundation, Inc.
 
   This file is part of GNU Inetutils.
 
@@ -1505,7 +1505,11 @@ setescape (int argc, char *argv[])
   else
     {
       printf ("new escape character: ");
-      fgets (buf, sizeof (buf), stdin);
+      if (fgets (buf, sizeof (buf), stdin) == NULL)
+	{
+	  buf[0] = '\0';
+	  printf ("\n");
+	}
       arg = buf;
     }
   if (arg[0] != '\0')
@@ -2472,10 +2476,15 @@ tn (int argc, char *argv[])
     {
       strcpy (line, "open ");
       printf ("(to) ");
-      fgets (&line[strlen (line)], sizeof (line) - strlen (line), stdin);
-      makeargv ();
-      argc = margc;
-      argv = margv;
+      if (fgets (&line[strlen (line)],
+		 sizeof (line) - strlen (line), stdin))
+	{
+	  makeargv ();
+	  argc = margc;
+	  argv = margv;
+	}
+      else
+	printf ("?Name of host was not understood.\n");
     }
   cmd = *argv;
   --argc;
@@ -2805,11 +2814,15 @@ tn (int argc, char *argv[])
       env_export ("USER");
     }
   call (status, "status", "notmuch", 0);
+  err = 0;
   if (setjmp (peerdied) == 0)
     telnet (user);
+  else
+    err = 1;
 
   close (net);
-  ExitString ("Connection closed by foreign host.\n", 1);
+  ExitString ("Connection closed by foreign host.\n", err);
+  /* NOT REACHED */
   return 0;
 }
 
@@ -2933,14 +2946,14 @@ command (int top, char *tbuf, int cnt)
   if (!top)
     {
       putchar ('\n');
-#if defined unix || defined __unix || defined __unix__
     }
+#if defined unix || defined __unix || defined __unix__
   else
     {
       signal (SIGINT, SIG_DFL);
       signal (SIGQUIT, SIG_DFL);
-#endif /* unix || __unix || __unix__ */
     }
+#endif /* unix || __unix || __unix__ */
   for (;;)
     {
       if (rlogin == _POSIX_VDISABLE)
@@ -2967,6 +2980,7 @@ command (int top, char *tbuf, int cnt)
 	    {
 	      if (feof (stdin) || ferror (stdin))
 		{
+		  printf ("\n");
 		  quit ();
 		}
 	      break;
