@@ -1,7 +1,7 @@
 /*
   Copyright (C) 1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003,
-  2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Free Software
-  Foundation, Inc.
+  2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014,
+  2015 Free Software Foundation, Inc.
 
   This file is part of GNU Inetutils.
 
@@ -95,7 +95,7 @@ init_network (void)
 int
 stilloob (void)
 {
-  static struct timeval timeout = { 0 };
+  static struct timeval timeout = { 0, 0 };
   fd_set excepts;
   int value;
 
